package eu.mhsl.minecraft.pixelblocks.pixelblock;

import co.aikar.taskchain.TaskChain;
import eu.mhsl.minecraft.pixelblocks.PixelBlockItem;
import eu.mhsl.minecraft.pixelblocks.Main;
import eu.mhsl.minecraft.pixelblocks.utils.Direction;
import net.kyori.adventure.text.Component;
import net.kyori.adventure.text.format.NamedTextColor;
import org.bukkit.*;
import org.bukkit.block.data.BlockData;
import org.bukkit.entity.*;
import org.bukkit.util.Vector;
import org.jetbrains.annotations.NotNull;

import javax.annotation.Nullable;
import java.io.File;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

public class PixelBlock {
    private boolean exists = true;
    private PixelBlockWorld pixelWorld;

    private final int pixelsPerBlock = Main.configuration.pixelsPerBlock();

    private final Location pixelBlockLocation;
    private final Direction facingDirection;
    private List<Pixel> pixels = new ArrayList<>();

    private PixelBlockHitbox hitbox;
    private PixelBlockPlaceholder placeholder;

    private Location lastEntryLocation;
    private final UUID ownerUUID;
    private final UUID blockUUID;

    public static final int maxPixelsPerBlock = 2000;

    public static @NotNull String getWorldName(@NotNull PixelBlock pixelBlock) {
        return Main.plugin.getDataFolder().getPath() + File.separator + pixelBlock.blockUUID;
    }

    public static @Nullable PixelBlock getPixelBlockFromBlockWorld(World world) {
        return Main.pixelBlocks.stream()
                .filter(block -> block.blockUUID.equals(getUUIDFromWorld(world)))
                .findFirst()
                .orElse(null);
    }

    public static @Nullable UUID getUUIDFromWorld(@NotNull World world) {
        try {
            return UUID.fromString(List.of(world.getName().split("/")).getLast());
        } catch (IllegalArgumentException e) {
            return null;
        }
    }

    public static @Nullable PixelBlock getPixelBlockFromPlacedLocation(@NotNull Location placedLocation) {
        Location searchLocation = placedLocation.clone().toBlockLocation();
        searchLocation.setPitch(0);
        searchLocation.setYaw(0);

        return Main.pixelBlocks.stream()
                .filter(block -> Objects.equals(block.pixelBlockLocation, searchLocation))
                .findFirst()
                .orElse(null);
    }

    public static PixelBlock fromExisting(UUID blockUUID, UUID ownerUUID, Location pixelBlockLocation, Direction direction, Location lastEntryLocation) {
        return new PixelBlock(blockUUID, ownerUUID, pixelBlockLocation, direction, lastEntryLocation);
    }
    private PixelBlock(UUID blockUUID, UUID ownerUUID, Location pixelBlockLocation, Direction direction, Location lastEntryLocation) {
        this.blockUUID = blockUUID;
        this.ownerUUID = ownerUUID;
        this.pixelBlockLocation = pixelBlockLocation;
        this.pixelBlockLocation.setYaw(0);
        this.pixelBlockLocation.setPitch(0);
        this.facingDirection = direction;
        this.lastEntryLocation = lastEntryLocation;

        this.pixelWorld = new PixelBlockWorld(this);
        this.pixels = Pixel.fromExisting(this);
        this.placeholder = PixelBlockPlaceholder.fromExisting(this);
        this.hitbox = PixelBlockHitbox.fromExisting(this);
    }

    public static PixelBlock createPixelBlock(UUID blockUUID, UUID ownerUUID, Location pixelBlockLocation, Direction direction) {
        return new PixelBlock(blockUUID, ownerUUID, pixelBlockLocation, direction);
    }
    private PixelBlock(UUID blockUUID, UUID ownerUUID, Location pixelBlockLocation, Direction direction) {
        if(Main.pixelBlocks.stream().anyMatch(pixelBlock -> pixelBlock.getBlockUUID().equals(blockUUID)))
            throw new IllegalStateException(String.format("PixelBlock '%s' ist bereits in der Welt vorhanden!", blockUUID));

        this.blockUUID = blockUUID;
        this.ownerUUID = ownerUUID;
        this.pixelBlockLocation = pixelBlockLocation;
        this.pixelBlockLocation.setYaw(0);
        this.pixelBlockLocation.setPitch(0);
        this.facingDirection = direction;

        this.getBlockTaskChain()
            .sync(() -> {
                this.pixelWorld = new PixelBlockWorld(this);
                this.placeholder = PixelBlockPlaceholder.newPlaceholder(this);
                this.hitbox = PixelBlockHitbox.newHitbox(this);
            })
            .execute();

        this.scheduleEntityUpdate();

        this.getBlockTaskChain()
            .async(() -> {
                Main.database.savePixelBlock(this);
                Main.pixelBlocks.add(this);
            })
            .execute();
    }

    public <T> TaskChain<T> getBlockTaskChain() {
        return Main.sharedChain(this.blockUUID.toString());
    }

    public void enterBlock(@NotNull Player player) {
        if(Main.configuration.onlyEditableByOwner() && !player.getUniqueId().equals(ownerUUID)) {
            player.sendMessage(Component.text("Dieser Pixelblock gehört nicht dir!", NamedTextColor.RED));
            return;
        }

        getBlockTaskChain()
            .async(() -> {
                this.lastEntryLocation = player.getLocation();
                Main.database.savePixelBlock(this);
            })
            .sync(() -> {
                if(!exists) return;
                player.teleport(this.pixelWorld.getSpawnLocation());
            })
            .current(() -> Main.plugin.getLogger()
                .info(String.format("'%s' entered PixelBlock '%s' at %s", player.getName(), this.blockUUID, this.pixelBlockLocation.toString())))
            .execute();
    }

    public void exitBlock(@NotNull Player player) {
        this.getBlockTaskChain()
            .sync(() -> player.teleport(this.lastEntryLocation != null ? this.lastEntryLocation : this.pixelBlockLocation))
            .current(() -> Main.plugin.getLogger().info(String.format("%s exited PixelBlock", player.getName())))
            .execute();

        this.scheduleEntityUpdate();
    }

    private void scheduleEntityUpdate() {
        record PixelData(PixelBlock parent, Vector relativeLocation, BlockData block, double scale) {
            Pixel create() {
                return Pixel.newPixel(this.parent, this.relativeLocation, this.block, this.scale);
            }
        }
        List<PixelData> pixelData = new ArrayList<>();

        this.scheduleEntityRemove();
        this.getBlockTaskChain()
            .async(() -> {
                for (int x = 0; x < pixelsPerBlock; x++) {
                    for (int y = 0; y < pixelsPerBlock; y++) {
                        for (int z = 0; z < pixelsPerBlock; z++) {
                            World world = pixelBlockLocation.getWorld();
                            Location relativeLocation = new Location(world, x, y, z);

                            Location blockLocation = this.pixelWorld.getBuildOrigin();
                            switch (this.facingDirection) {
                                case south -> blockLocation.add(relativeLocation.x(), relativeLocation.y(), relativeLocation.z());
                                case north -> blockLocation.add((pixelsPerBlock-1)-relativeLocation.x(), relativeLocation.y(), (pixelsPerBlock-1)-relativeLocation.z());
                                case east -> blockLocation.add((pixelsPerBlock-1)-relativeLocation.z(), relativeLocation.y(), relativeLocation.x());
                                case west -> blockLocation.add(relativeLocation.z(), relativeLocation.y(), (pixelsPerBlock-1)-relativeLocation.x());
                            }
                            BlockData block = blockLocation.getBlock().getBlockData();

                            if(!block.getMaterial().isEmpty()) {
                                pixelData.add(new PixelData(this, relativeLocation.toVector(), block, (double) 1 / pixelsPerBlock));
                            }
                        }
                    }
                }
            })
            .async(() -> Collections.shuffle(pixelData))
            .asyncFirst(() -> {
                int spawnSpreadInTicks = 10;
                int chunkSize = (int) Math.ceil((double) pixelData.size() / spawnSpreadInTicks);

                TaskChain<Object> spawnTask = this.getBlockTaskChain();
                IntStream.range(0, spawnSpreadInTicks)
                    .mapToObj(i -> pixelData.stream()
                        .skip((long) i * chunkSize)
                        .limit(chunkSize)
                        .collect(Collectors.toList()))
                    .forEach(pixels -> spawnTask
                        .delay(1)
                        .sync(() -> pixels.forEach(pixel -> this.pixels.add(pixel.create()))));

                return spawnTask;
            })
            .syncLast((chain) -> chain
                .sync(() -> {
                    this.hitbox = PixelBlockHitbox.newHitbox(this);
                    this.placeholder = PixelBlockPlaceholder.newPlaceholder(this);
                    Main.plugin.getLogger().info(String.format("Placed %d entities for PixelBlock '%s'", this.pixels.size(), this.blockUUID));
                })
                .execute())
            .execute();
    }

//    public void place(Location placeLocation, Direction direction) {
//        Location newLocation = placeLocation.toBlockLocation();
//        newLocation.setPitch(0);
//        newLocation.setYaw(0);
//
//        @Nullable PixelBlock blockAtLocation = PixelBlock.getPixelBlockFromPlacedLocation(newLocation);
//        if(blockAtLocation != null && blockAtLocation != this) throw new IllegalArgumentException("Es können nicht mehrere Pixelblöcke ineinander platziert werden.");
//
//        Main.plugin.getLogger().info(String.format("Placing PixelBlock '%s' at %s", this.blockUUID, placeLocation));
//        this.pixelBlockLocation = newLocation;
//        this.facingDirection = direction;
//        updateEntities();
//        Main.database.savePixelBlock(this);
//
//        this.placeholder = PixelBlockPlaceholder.newPlaceholder(this);
//    }

    public void destroy(Player destroyedBy) {
        if(Main.configuration.onlyBreakableByOwner() && !destroyedBy.getUniqueId().equals(ownerUUID)) {
            destroyedBy.sendMessage("Dieser Pixelblock gehört nicht dir!");
            return;
        }

        this.pixelWorld.getPlayersInWorld().forEach(p -> {
            p.sendMessage(Component.text("Der Pixelblock wurde von einem anderen Spieler abgebaut!", NamedTextColor.RED));
            p.teleport(this.lastEntryLocation);
        });

        Main.plugin.getLogger().info(String.format("Destroying PixelBlock '%s' at %s", this.blockUUID, pixelBlockLocation));
        this.exists = false;

        this.pixelWorld.getEntitiesInWorld().stream()
            .filter(entity -> entity instanceof Item)
            .forEach(entity -> entity.teleport(this.lastEntryLocation));

        this.scheduleEntityRemove();
        this.getBlockTaskChain()
            .sync(() -> {
                World world = this.pixelBlockLocation.getWorld();
                world.playSound(this.pixelBlockLocation, Sound.BLOCK_COPPER_BULB_BREAK, 1.0F, 30);
                world.dropItem(this.pixelBlockLocation.add(new Vector(0.5, 0.5, 0.5)), PixelBlockItem.getBlockAsItem(this));
            })
            .async(() -> {
                Main.database.deletePixelBlock(this);
                Main.pixelBlocks.remove(this);
            })
            .execute();
    }

    private void scheduleEntityRemove() {
//        Chunk chunk = this.pixelBlockLocation.getChunk();
//        if(!chunk.isEntitiesLoaded()) {
//            chunk.load(true);
//            chunk.getEntities();
//        }

        this.getBlockTaskChain()
            .current(() -> Main.plugin.getLogger()
                .info(String.format("Removing %d entities for PixelBlock '%s'", this.pixels.size(), this.blockUUID)))
            .sync(() -> {
                this.pixels.forEach(Pixel::destroy);
                this.pixels.clear();
                this.placeholder.destroy();
                this.hitbox.destroy();
            })
            .execute();
    }

    public @NotNull PixelBlockWorld getPixelWorld() {
        return pixelWorld;
    }

    public Location getPixelBlockLocation() {
        return pixelBlockLocation.clone();
    }

    public Direction getFacingDirection() {
        return facingDirection;
    }

    public boolean hasLastEntryLocation() {
        return this.lastEntryLocation != null;
    }

    public List<Pixel> getPixels() {
        return pixels;
    }

    public Location getLastEntryLocation() {
        return lastEntryLocation.clone();
    }

    public UUID getOwnerUUID() {
        return ownerUUID;
    }

    public UUID getBlockUUID() {
        return blockUUID;
    }
}
